# Interfaces.pm - Settings file for generator (IPTables Firewall Configuration).
# Created by James A. Pattie.  Copyright (c) 2001-2004, Xperience, Inc.

package PCXFireWall::Interfaces;

use strict;
use vars qw ($AUTOLOAD @ISA @EXPORT $VERSION);

require Exporter;
@ISA = qw(Exporter AutoLoader);
@EXPORT = qw();

$VERSION = "2.23";

sub new
{
  my $that = shift;
  my $class = ref($that) || $that;
  my $self = bless {}, $class;
  $self->{interfaces} = {};

  # populate with the predefined locations
  foreach my $location ("internal", "external", "dmz", "firewall")
  {
    $self->{interfaces}->{$location} = {};
  }

  return $self;
}

sub AUTOLOAD
{
  my $self = shift;
  my $type = ref($self) || die "$self is not an object";
  my $name = $AUTOLOAD;
  $name =~ s/.*://; # strip fully-qualified portion
  unless (exists $self->{$name})
  {
    die "Can't access `$name' field in object of class $type";
  }
  if (@_)
  {
    return $self->{$name} = shift;
  }
  else
  {
    return $self->{$name};
  }
}

# defineInterface
# requires: name (ethX), type (static|dynamic), location (internal|external|dmz|user defined value),
#           alias - user defined name for the interface (ex. 'external', 'vpn0', etc.)
# optional: IP (IP Address)[required if type = static], dynamicIPMode (ignoreIP, lookupIP)
# returns: nothing
# summary: This creates an entry in the interfaces hash that defines an internal or external
# network card and it's settings.
sub defineInterface
{
  my $self = shift;
  my %args = ( name => "eth0", type => "static", location => "external", IP => "", alias => "", dynamicIPMode => "lookupIP", @_ );
  my $name = $args{name};
  my $type = $args{type};
  my $location = $args{location};
  my $IP = $args{IP};
  my $alias = $args{alias};
  my $dynamicIPMode = $args{dynamicIPMode};

  if (length $name == 0)
  {
    die "(defineInterface) - Error:  You must specify the name for the interface!\n";
  }
  if (length $alias == 0)
  {
    ($alias = $name) =~ s/^([^:]+)(:.+)$/$1/;  # set equal to the name, blow a warning.
    print "WARNING:  You did not specify the alias, so I'm setting it to '$alias'!\n";
  }
  else
  {
    $alias =~ s/^([^:]+)(:.+)$/$1/;  # make sure there is not a :x part in it.
  }
  if ($type !~ /^(static|dynamic)$/)
  {
    die "(defineInterface) - Error:  type = '$type' is invalid!\n";
  }
  if (!exists $self->{interfaces}->{$location})
  {
    die "(defineInterface) - Error:  location = '$location' is invalid!\n";
  }
  if ($type eq "static")
  {  # validate the IP address given.
    if (length $IP == 0)
    {
      die "(defineInterface) - Error:  IP must be specified when type = 'static'!\n";
    }
    if ($IP !~ m/^([01]?\d\d|2[0-4]\d|25[0-5]|\d)\.([01]?\d\d|2[0-4]\d|25[0-5]|\d)\.([01]?\d\d|2[0-4]\d|25[0-5]|\d)\.([01]?\d\d|2[0-4]\d|25[0-5]|\d)$/)
    {
      die "(defineInterface) - Error: IP = '$IP' is invalid!\n";
    }
  }
  else
  {
    if ($dynamicIPMode !~ /^(lookupIP|ignoreIP)$/)
    {
      die "(defineInterface) - Error: dynamicIPMode = '$dynamicIPMode' is invalid!\n";
    }
  }

  # make sure we haven't already created this interface definition.
  if (exists $self->{interfaces}->{$location}->{$name})
  {
    die "(defineInterface) - Error:  You have already defined '$name' as an $location interface!\n";
  }
  # make sure the specified alias is not already being used.
  foreach my $location (keys %{$self->{interfaces}})
  {
    foreach my $interface (keys %{$self->{interfaces}->{$location}})
    {
      if ($self->{interfaces}->{$location}->{$interface}->{alias} eq $alias)
      {
        die "(defineInterface) - Error: alias = '$alias' already defined for $name as an $location interface!\n";
      }
    }
  }

  # now create the entry in the interfaces hash.
  $self->{interfaces}->{$location}->{$name}->{type} = $type;
  $self->{interfaces}->{$location}->{$name}->{IP} = $IP;
  $self->{interfaces}->{$location}->{$name}->{usingBroadcast} = 0;  # This flags lets the generator know if it needs to make the variable.
  $self->{interfaces}->{$location}->{$name}->{alias} = $alias;  # this is what the shell script will look for and the rules also, now.
  $self->{interfaces}->{$location}->{$name}->{dynamicIPMode} = $dynamicIPMode;  # remember what mode the user specified.
}

# defineLocation
# requires: location
# optional:
# returns: nothing
# summary: This routine allows you to create alternate 'locations' that
#    your network interfaces can be grouped under than the default
#    locations of internal, external, firewall and dmz.
sub defineLocation
{
  my $self = shift;
  my %args = ( location => "", @_ );
  my $location = $args{location};

  if (length $location == 0)
  {
    die "(defineLocation) - Error:  You must specify the location to create!\n";
  }
  if ($location =~ /^(internal|external|dmz|firewall)$/)
  {
    die "(defineLocation) - Error:  You can not redeclare an internal location!  location = '$location'.\n";
  }
  if (exists $self->{interfaces}->{$location})
  {
    die "(defineLocation) - Error:  location = '$location' already created!\n";
  }

  $self->{interfaces}->{$location} = {};  # define the new location.
}

# getInterfacesLocatedAt
# requires: location (internal|external|dmz|user defined value)
# returns: hash of all interfaces of the specified location.
# summary: This routine validates the specified location and then returns
#          the hash that contains all interfaces defined for that location.
sub getInterfacesLocatedAt
{
  my $self = shift;
  my %args = ( location => "external", @_ );
  my $location = $args{location};
  my %result;

  if (!exists $self->{interfaces}->{$location})
  {
    die "(getInterfacesLocatedAt) - Error: location = '$location' is invalid!\n";
  }

  %result = %{$self->{interfaces}->{$location}};

  return %result;
}

# @(location, type, ip, usingBroadcast, alias) getInfo(int)
#    This routine takes an interface the user has registered
#    and looks it up.  If it is not found, then an empty array
#    is returned, else the location, type and ip info is returned as
#    depicted in the return value.
sub getInfo
{
  my $self = shift;
  my %args = ( int => "", @_ );
  my $int = $args{int};
  my @result = ();

  if (length $int == 0)
  {
    die "(getInfo) - Error:  You must specify the interface to get info on!\n";
  }

  foreach my $location (keys %{$self->{interfaces}})
  {
    foreach my $interface (keys %{$self->{interfaces}->{$location}})
    {
      if ($interface eq $int)
      {
        $result[0] = $location;
        $result[1] = $self->{interfaces}->{$location}->{$int}->{type};
        $result[2] = $self->{interfaces}->{$location}->{$int}->{IP};
        $result[3] = $self->{interfaces}->{$location}->{$int}->{usingBroadcast};
        $result[4] = $self->{interfaces}->{$location}->{$int}->{alias};
        $result[5] = $self->{interfaces}->{$location}->{$int}->{dynamicIPMode};
      }
    }
  }

  return @result;
}

# void usingBroadcast(int)
# requires: int - interface to work with
# returns: nothing
# summary:  finds the specified interface and sets the usingBroadcast flag to 1
sub usingBroadcast
{
  my $self = shift;
  my %args = ( int => "", @_ );
  my $int = $args{int};

  if (length $int == 0)
  {
    die "(usingBroadcast) - Error:  You must specify the interface to work with!\n";
  }

  my $found = 0;
  foreach my $location (keys %{$self->{interfaces}})
  {
    foreach my $interface (keys %{$self->{interfaces}->{$location}})
    {
      if ($interface eq $int)
      {
        $self->{interfaces}->{$location}->{$int}->{usingBroadcast} = 1;
        $found = 1;
        last;
      }
    }
  }
  if (!$found)
  {
    die "(usingBroadcast) - Error:  Could not find interface = '$int'!\n";
  }
}

# scalar displayInterfacesAndIPs()
# requires: nothing
# returns: string
# summary: outputs to the string all Interfaces and their IP's and
#          BROADCAST variables.
sub displayInterfacesAndIPs
{
  my $self = shift;
  my %args = ( @_ );
  my $result = "";

  foreach my $location (keys %{$self->{interfaces}})
  {
    foreach my $name (keys %{$self->{interfaces}->{$location}})
    {
      my $type = $self->{interfaces}->{$location}->{$name}->{type};
      my $alias = $self->{interfaces}->{$location}->{$name}->{alias};
      my $UCalias = uc($alias);
      my $ip = $self->{interfaces}->{$location}->{$name}->{IP};
      (my $int = $name) =~ s/^([^:]+)(:.+)$/$1/;  # strip off any :x stuff from the interface name.

      # generate the INT_alias variable.
      $result .= "  INT_$UCalias=$int\n";

      if ($type eq "dynamic")
      {
        if ($self->{interfaces}->{$location}->{$name}->{dynamicIPMode} eq "lookupIP")
        {
          $result .= "  " . $UCalias . "_IP=`getIPAddress $name $alias`\n";
        }
      }
      else
      {
        $result .= "  " . $UCalias . "_IP=\"$ip\"\n";
      }
      my $usingBroadcast = $self->{interfaces}->{$location}->{$name}->{usingBroadcast};
      # define the $<alias>BROADCAST shell variable
      if (($type eq "dynamic" && $self->{interfaces}->{$location}->{$name}->{dynamicIPMode} eq "lookupIP") || $type eq "static")
      {
        $result .= "  " . $UCalias . "_BROADCAST=`getBroadCastAddress $name $alias`\n";
      }

      # seperator
      $result .= "\n";
    }
  }

  return $result;
}

# bool dynamicInterfaces()
# requires: nothing
# returns: nothing
# summary: returns 1 if there are any dynamic interfaces, else 0.
sub dynamicInterfaces
{
  my $self = shift;
  my %args = ( @_ );
  my $result = 0;

  foreach my $location (keys %{$self->{interfaces}})
  {
    foreach my $name (keys %{$self->{interfaces}->{$location}})
    {
      my $type = $self->{interfaces}->{$location}->{$name}->{type};
      if ($type eq "dynamic")
      {
        $result = 1;
      }
    }
  }

  return $result;
}

# scalar displayShellFunctions()
#   This routine will output the necessary shell functions as a string.
#   This is called by the generator script.
sub displayShellFunctions
{
  my $self = shift;
  my %args = ( @_ );
  my $result = "";

  $result .= <<"END_OF_CODE";

# \$1 = interface, \$2 = alias
function getIPAddress ()
{
  my_IP=""
  if [ "\$NETWORK_COMMAND" == "ifconfig" ]; then
    my_IP=`\$NETCOMMAND \$1 | grep inet | grep -v inet6 | cut -d ":" -f 2 | cut -d " " -f 1`
  else
    if [ "\$NETWORK_COMMAND" == "ip" ]; then
      # check for a :x interface
      alias=`echo \$1 | grep :`
      if [ ! -z "\$alias" ]; then
        # have to strip off the :x part
        interface=`echo \$1 | cut -d ":" -f 1`
        my_IP=`\$NETCOMMAND addr show dev \$interface | grep \$1 | grep inet | grep -v inet6 | cut -d " " -f 6 | cut -d "/" -f 1`
      else
        interface=\$1
        my_IP=`\$NETCOMMAND addr show dev \$interface | grep inet | grep -v inet6 | cut -d " " -f 6 | cut -d "/" -f 1`
      fi
    else
      echo "'\$NETWORK_COMMAND' is unknown command!" >&2
      exit 1
    fi
  fi

  # now make sure we got an IP Address.
  if [ -z "\$my_IP" ]; then
    echo \"No IP Address returned for interface \$2 = \$1!" >&2
    exit 1
  fi
  echo \$my_IP
}

# \$1 = interface, \$2 = alias
function getBroadCastAddress ()
{
  my_BroadCastIP=""
  if [ "\$NETWORK_COMMAND" == "ifconfig" ]; then
    my_BroadCastIP=`\$NETCOMMAND \$1 | grep Bcast | cut -d ":" -f 3 | cut -d " " -f 1`
  else
    if [ "\$NETWORK_COMMAND" == "ip" ]; then
      # check for a :x interface
      alias=`echo \$1 | grep :`
      if [ ! -z "\$alias" ]; then
        # have to strip off the :x part
        interface=`echo \$1 | cut -d ":" -f 1`
        my_BroadCastIP=`\$NETCOMMAND addr show dev \$interface | grep \$1 | grep inet | grep -v inet6 | cut -d " " -f 8`
      else
        interface=\$1
        my_BroadCastIP=`\$NETCOMMAND addr show dev \$interface | grep inet | grep -v inet6 | grep -v secondary | cut -d " " -f 8`
      fi
    else
      echo "'\$NETWORK_COMMAND' is unknown command!" >&2
      exit 1
    fi
  fi

  # now make sure we got a BroadCast Address.
  if [ -z "\$my_BroadCastIP" ]; then
    echo \"No BroadCast Address returned for interface \$2 = \$1!" >&2
    exit 1
  fi
  echo \$my_BroadCastIP
}

END_OF_CODE

  return $result;
}

1;
__END__

=head1 NAME

Interfaces - The PCXFireWall Interface Definition Object.

=head1 SYNOPSIS

  use PCXFireWall::Interfaces;
  my $obj = PCXFireWall::Interfaces->new;

  # Define the internal and external Network Interfaces being
  # used/protected.
  $obj->defineInterface(name => "ppp0", alias => "dialup0",
                        type => "dynamic");
  $obj->defineInterface(name => "eth0", alias => "internal0",
                        location => "internal", IP => "192.168.0.250");
  $obj->defineInterface(name => "eth0:0", alias => "internal1",
                        location => "internal", IP => "192.168.0.251");
  $obj->defineInterface(name => "eth1", alias => "dmz0",
                        location => "dmz", IP => "192.168.1.250");

  # define a new location - vpn
  $obj->defineLocation(location => "vpn");

  # get all dmz interfaces
  my %dmzs = $obj->getInterfacesLocatedAt(location => "dmz");
  foreach my $interface (keys %dmzs)
  {
    print "Interface '$interface', type = '$dmzs{$interface}{type}'," .
    "alias = '$dmzs{$interface}{alias}', IP = '$dmzs{$interface}{IP}'.\n";
  }

=head1 DESCRIPTION

Interfaces is the Interface Definition Object for the PCXFireWall.

=head1 FUNCTIONS

  scalar new(void)
    Creates a new instance of the PCXFireWall::Interfaces object.

  void defineInterface(name, alias, type, location, IP, dynamicIPMode)
    requires: name (ethX), type (static|dynamic),
              location (internal|external|dmz|firewall|user defined value)
              alias (name used by the shell script)
    optional: IP (IP Address)[required if type = static],
              dynamicIPMode (lookupIP or ignoreIP) - only valid when type
              = 'dynamic'.  lookupIP implements the historical code of
              using the interface_IP variable.  ignoreIP causes rules to
              be generated that do not use the interface_IP variable.
              The default, when type = dynamic, is lookupIP.
    returns: nothing
    summary: This creates an entry in the interfaces hash that defines
             an internal or external network card and it's settings.
             The type defaults to static and location defaults to
             external, so you don't have to specify them if that is
             what you need.

  void defineLocation(location)
    requires: location
    optional:
    returns: nothing
    summary: This routine allows you to create alternate 'locations' that
             your network interfaces can be grouped under than the
             default locations of internal, external, firewall and dmz.

  %interfaces getInterfacesLocatedAt(location)
    requires: location (internal|external|dmz|user defined value)
    returns: hash of all interfaces of the specified location.
    summary: This routine validates the specified location and then
             returns the hash that contains all interfaces defined for
             that location.

  @(location, type, ip, usingBroadcast, alias) getInfo(int)
    This routine takes an interface the user has registered
    and looks it up.  If it is not found, then an empty array
    is returned, else the location, type and ip info is returned as
    depicted in the return value.

  void usingBroadcast(int)
    requires: int - interface to work with
    returns: nothing
    summary:  finds the specified interface and sets the usingBroadcast
              flag to 1

  scalar displayInterfacesAndIPs()
    requires: nothing
    returns: nothing
    summary: outputs to the string all Interfaces and their IP's and
             BROADCAST variables.

  scalar displayShellFunctions()
    This routine will output the necessary shell functions as a string.
    This is called by the generator script.

  bool dynamicInterfaces()
    requires: nothing
    returns: nothing
    summary: returns 1 if there are any dynamic interfaces, else 0.

=head1 VARIABLES

  interfaces - hash of all interfaces defined by the user.
               Format is
                interfaces->{location}->{interface name}->{type}
                interfaces->{location}->{interface name}->{IP}
                interfaces->{location}->{interface name}->{usingBroadcast}
                interfaces->{location}->{interface name}->{alias}

  NOTE:  All data fields are accessible by specifying the object
         and pointing to the data member to be modified on the
         left-hand side of the assignment.
         Ex.  $obj->variable($newValue); or $value = $obj->variable;

=head1 AUTHOR

Xperience, Inc. (mailto:admin at pcxperience.com)

=head1 SEE ALSO

perl(1)

=cut
