# Modules.pm - Interface module for insmod support for generator (IPTables Firewall Configuration).
# Created by James A. Pattie.  Copyright (c) 2001-2004, Xperience, Inc.

package PCXFireWall::Modules;

use strict;
use vars qw ($AUTOLOAD @ISA @EXPORT $VERSION);

require Exporter;
@ISA = qw(Exporter AutoLoader);
@EXPORT = qw();

$VERSION = "2.23";

sub new
{
  my $that = shift;
  my $class = ref($that) || $that;
  my $self = bless {}, $class;
  $self->{modules} = {};
  $self->{modules}->{'_order'} = [];  # the order in which to output.
  $self->{kernelType} = "monolithic";

  $self->register(module => "ip_tables", params => "");
  $self->register(module => "iptable_filter", params => "");
  $self->register(module => "ipt_LOG", params => "");

  return $self;
}

sub AUTOLOAD
{
  my $self = shift;
  my $type = ref($self) || die "$self is not an object";
  my $name = $AUTOLOAD;
  $name =~ s/.*://;	# strip fully-qualified portion
  unless (exists $self->{$name})
  {
    die "Can't access `$name' field in object of class $type";
  }
  if (@_)
  {
    return $self->{$name} = shift;
  }
  else
  {
    return $self->{$name};
  }
}

# register(module, params)
#  required: module
#  optional: params
#  summary:  This makes an entry in the modules hash which specifies
#            the modules that need to be checked for and any options
#            they need to be insmod'ed with.
sub register
{
  my $self = shift;
  my %args = ( module => "", params => "", @_ );
  my $module = $args{module};
  my $params = $args{params};
  my $errStr = "PCXFireWall::Modules->register()  - Error!\n";

  if (length $module == 0)
  {
    die $errStr . "module must be specified!\n";
  }

  if (exists $self->{modules}->{$module})
  {
    die $errStr . "module = '$module' already defined!\n";
  }

  my $code = <<"END_OF_CODE";

    # check for module $module.o or $module.ko
    checkForModule "$module" "$params"
END_OF_CODE

  push @{$self->{modules}->{'_order'}}, $module;
  $self->{modules}->{$module} = $code;
}

# void kernelType(type)
# required: type - monolithic or modular
# summary: specifies the type of kernel and whether or not the
#          code needs to be generated when display() is called.
sub kernelType
{
  my $self = shift;
  my %args = ( type => "monolithic", @_ );
  my $type = $args{type};

  if ($type !~ /^(monolithic|modular)$/)
  {
    die "(Modules->kernelType) - Error: type = '$type' is invalid!\n";
  }

  $self->{kernelType} = $type;
}

# scalar display()
#   This routine will write the rules to a string.  This is
#   called by the generator script.
sub display
{
  my $self = shift;
  my %args = ( @_ );
  my $result = "";

  return $result if ($self->{kernelType} eq "monolithic");

  $result .= "\n  #\n  # netfilter insmod support.\n  #\n\n";

  $result .= <<"END_OF_CODE";
  # make sure that netfilter support is modular in this kernel.
  if [ -d /lib/modules/\${KERNEL}/kernel/net/ipv4/netfilter ]; then
END_OF_CODE

  # now output all entries we generated.
  foreach my $entry (@{$self->{modules}->{'_order'}})
  {
    $result .= $self->{modules}->{$entry};
  }

  $result .= "  fi\n"; #end of modular support if.
  
  return $result;
}

# scalar displayShellFunction()
#   This routine will output the necessary shell function to a string.
#   This is called by the generator script.
sub displayShellFunction
{
  my $self = shift;
  my %args = ( @_ );
  my $result = "";

  $result .= <<"END_OF_CODE";

KERNEL=`uname -r`

function checkForModule ()
{
  # check for module \$1 with params \$2
  if [ -e /lib/modules/\${KERNEL}/kernel/net/ipv4/netfilter/\$1.o ]; then
    OUTPUT=`/sbin/lsmod | grep \$1`
    if [ -z "\${OUTPUT}" ]; then
      /sbin/insmod /lib/modules/\${KERNEL}/kernel/net/ipv4/netfilter/\$1.o \$2
    fi
  else
    if [ -e /lib/modules/\${KERNEL}/kernel/net/ipv4/netfilter/\$1.ko ]; then
      OUTPUT=`/sbin/lsmod | grep \$1`
      if [ -z "\${OUTPUT}" ]; then
        /sbin/insmod /lib/modules/\${KERNEL}/kernel/net/ipv4/netfilter/\$1.ko \$2
      fi
    else
      echo "Error: module = '\$1' does not exist! kernel = \${KERNEL}"
    fi
  fi
}

END_OF_CODE

  return $result;
}

1;
__END__

=head1 NAME

Modules - The PCXFireWall netfilter insmod interface.

=head1 SYNOPSIS

  use PCXFireWall::Modules;
  my $modules = PCXFireWall::Modules->new;

  # specify that we need the conntrack module.
  $modules->register(module => "ip_conntrack");

=head1 DESCRIPTION

Modules is the netfilter insmod interface Object for the PCXFireWall.

=head1 FUNCTIONS

  scalar new(void)
    Creates a new instance of the PCXFireWall::Modules object.
    Initializes the modules hash.
    By default ip_tables, iptable_filter and ipt_LOG are specified
    to be loaded.

  void kernelType(type)
    type is either modular or monolithic.  If monolithic, then when
    display() is called, the output will not be generated.

  scalar display()
    This routine will write the rules to a string.  This is
    called by the generator script.

  scalar displayShellFunction()
    This routine will output the necessary shell function to a string.
    This is called by the generator script.

  void register(module, params)
    required: module
    optional: params
    summary:  This makes an entry in the modules hash which specifies
              the modules that need to be checked for and any options
              they need to be insmod'ed with.

=head1 VARIABLES

  modules - hash of modules that need to be checked for and insmoded
            if not present or compiled into the kernel.

  kernelType - keeps track of whether or not you want the rules
               generated.  By default is 'monolithic', which means no
               code will be generated to try and insert modules, etc.

  NOTE:  All data fields are accessible by specifying the object
         and pointing to the data member to be modified on the
         left-hand side of the assignment.
         Ex.  $obj->variable($newValue); or $value = $obj->variable;

=head1 AUTHOR

Xperience, Inc. (mailto:admin at pcxperience.com)

=head1 SEE ALSO

perl(1)

=cut
