# NAT.pm - IPTables NAT Module.
# Created by James A. Pattie.  Copyright (c) 2001-2004, Xperience, Inc.

package PCXFireWall::NAT;

use strict;
use vars qw ($AUTOLOAD @ISA @EXPORT $VERSION);

require Exporter;
@ISA = qw(Exporter AutoLoader);
@EXPORT = qw();

$VERSION = "2.23";

my $TABLE = " -t nat";  # This will hold the string " -t TABLE NAME" which defines the iptables table to work with.

# new
# requires: configObj, interfaceObj
sub new
{
  my $that = shift;
  my $class = ref($that) || $that;
  my $self = bless {}, $class;
  my %args = ( configObj => undef, interfaceObj => undef, @_ );
  my $configObj = $args{configObj};
  my $interfaceObj = $args{interfaceObj};

  if (!defined $configObj)
  {
    die "(NAT->new) - Error: configObj is not defined!\n";
  }
  $self->{configObj} = $configObj;

  if (!defined $interfaceObj)
  {
    die "(NAT->new) - Error: interfaceObj is not defined!\n";
  }
  $self->{interfaceObj} = $interfaceObj;

  $self->{userChains} = {}; # this hash will hold the name of each chain that is created by the user.
  $self->{rules} = [];  # array that each rule as it is defined is appended to.  This is the complete string that will be output to the file.
  $self->{setupRules} = []; # array to hold setup related rules in.

  $self->{orphanedChains} = {}; # This hash keeps track of userChains created and rules added to but never been jumped to.

  return $self;
}

sub AUTOLOAD
{
  my $self = shift;
  my $type = ref($self) || die "$self is not an object";
  my $name = $AUTOLOAD;
  $name =~ s/.*://; # strip fully-qualified portion
  unless (exists $self->{$name})
  {
    die "Can't access `$name' field in object of class $type";
  }
  if (@_)
  {
    return $self->{$name} = shift;
  }
  else
  {
    return $self->{$name};
  }
}

# void policy(chain, policy)
#    Sets the default policy of chain to policy.  Valid values are
#    PREROUTING, OUTPUT and POSTROUTING for chain and ACCEPT or DROP for policy.
sub policy
{
  my $self = shift;
  my %args = ( chain => "", policy => "", preComment => $self->{configObj}->{preComment}, postComment => $self->{configObj}->{postComment}, @_ );
  my $chain = $args{chain};
  my $policy = $args{policy};
  my $preComment = $args{preComment};
  my $postComment = $args{postComment};

  if ($chain !~ /^(PREROUTING|OUTPUT|POSTROUTING)$/)
  {
    die "(NAT->policy) - Error: chain = '$chain' is invalid!\n";
  }
  if ($policy !~ /^(ACCEPT|DROP)$/)
  {
    die "(NAT->policy) - Error: policy = '$policy' is invalid!\n";
  }

  my $rule = "  ";
  $rule .= $preComment . "  " if (length $preComment > 0);
  $rule .= "\$IPTABLES$TABLE -P $chain $policy\n";
  $rule .= "  " . $postComment if (length $postComment > 0);

  $self->appendRule(rule => $rule, mode => "setup");
}

# void createChain(name)
#    Makes an entry in userChains for the specified chain and then makes
#    an entry in rules for the command to create the chain.
sub createChain
{
  my $self = shift;
  my %args = ( name => "", preComment => $self->{configObj}->{preComment}, postComment => $self->{configObj}->{postComment}, @_ );
  my $name = $args{name};
  my $preComment = $args{preComment};
  my $postComment = $args{postComment};

  if (length $name == 0)
  {
    die "(NAT->createChain) - Error: name must be specified!\n";
  }
  if (length $name > 30)
  {
    die "(NAT->createChain) - Error: name = '$name' is too long!  It can only be 30 characters long.\n";
  }

  if ($name =~ /^(PREROUTING|OUTPUT|POSTROUTING)$/)
  {
    die "(NAT->createChain) - Error: You can not re-create a standard chain <$name>!\n";
  }

  if (exists $self->{userChains}->{$name})
  {
    die "(NAT->createChain) - Error: chain = '$name' already exists!\n";
  }
  $self->{userChains}->{$name} = -1;  # This is used by the rule methods to know if the chain has been jumped to yet.

  my $rule = "  ";
  $rule .= $preComment . "  " if (length $preComment > 0);
  $rule .= "\$IPTABLES$TABLE -N $name 2> /dev/null\n";
  $rule .= "  " . $postComment if (length $postComment > 0);

  $self->appendRule(rule => $rule, mode => "setup");
}

# void flushChain(name)
#    name can be 'All' which will empty all user defined chains.
#    Plus all Builtin chains.
#    A rule will be added which would flush the specified chain.
sub flushChain
{
  my $self = shift;
  my %args = ( name => "", preComment => $self->{configObj}->{preComment}, postComment => $self->{configObj}->{postComment}, @_ );
  my $name = $args{name};
  my $preComment = $args{preComment};
  my $postComment = $args{postComment};

  if (length $name == 0)
  {
    die "(NAT->flushChain) - Error: name must be specified!\n";
  }
  if (length $name > 30)
  {
    die "(NAT->flushChain) - Error: name = '$name' is too long!  It can only be 30 characters long.\n";
  }

  if ($name ne "All")
  {
    my $rule = "  ";
    $rule .= $preComment . "  " if (length $preComment > 0);
    $rule .= "\$IPTABLES$TABLE -F $name\n";
    $rule .= "  " . $postComment if (length $postComment > 0);

    $self->appendRule(rule => $rule, mode => "setup");
  }
  else # flush All Chains.
  {
    my $rule = "  ";
    $rule .= $preComment . "  " if (length $preComment > 0);
    $rule .= "\$IPTABLES$TABLE -F\n";
    $rule .= "  " . $postComment if (length $postComment > 0);

    $self->appendRule(rule => $rule, mode => "setup");
  }
}

# void deleteChain(name)
#    name can be 'All' which will delete all user defined chains.
#    The chain(s) will be removed from userChains (if present) and then
#    a rule(s) will be added which would delete the chain.  The chain(s)
#    will have rules created to flush them before we delete it so that
#    it will always work.  Thus you don't need to call flushChain unless
#    you are not planning on deleting the chain.
sub deleteChain
{
  my $self = shift;
  my %args = ( name => "", preComment => $self->{configObj}->{preComment}, postComment => $self->{configObj}->{postComment}, @_ );
  my $name = $args{name};
  my $preComment = $args{preComment};
  my $postComment = $args{postComment};

  if (length $name == 0)
  {
    die "(NAT->deleteChain) - Error: name must be specified!\n";
  }
  if (length $name > 30)
  {
    die "(NAT->deleteChain) - Error: name = '$name' is too long!  It can only be 30 characters long.\n";
  }

  if ($name =~ /^(PREROUTING|OUTPUT|POSTROUTING)$/)
  {
    die "(NAT->deleteChain) - Error: You can not delete a standard chain <$name>!\n";
  }

  if ($name ne "All")
  {
    $self->flushChain(name => $name);  # flush the chain first.
    if (exists $self->{userChains}->{$name})
    {
      delete $self->{userChains}->{$name};
    }

    delete $self->{orphanedChains}->{$name} if (exists $self->{orphanedChains}->{$name});

    my $rule = "  ";
    $rule .= $preComment . "  " if (length $preComment > 0);
    $rule .= "\$IPTABLES$TABLE -X $name\n";
    $rule .= "  " . $postComment if (length $postComment > 0);

    $self->appendRule(rule => $rule, mode => "setup");
  }
  else
  {
    $self->flushChain(name => $name); # flush all chains first.
    $self->{userChains} = {};  # empty the userChains hash.

    my $rule = "  ";
    $rule .= $preComment . "  " if (length $preComment > 0);
    $rule .= "\$IPTABLES$TABLE -X\n";   # delete all user defined chains.
    $rule .= "  " . $postComment if (length $postComment > 0);

    $self->appendRule(rule => $rule, mode => "setup");
  }
}

# scalar getBaseChain(chain, ruleName)
#    This routine will determine the base chain for the specified chain.
sub getBaseChain
{
  my $self = shift;
  my %args = ( chain => "", ruleName => "getBaseChain", @_ );
  my $chain = $args{chain};
  my $ruleName = $args{ruleName};
  my $baseChain = "";

  if (length $chain == 0)
  {
    die "(NAT->getBaseChain) - Error:  You must specify the chain to work with!\n";
  }
  # find out what our base rule is (PREROUTING, OUTPUT, POSTROUTING, -1 [not referenced yet]);
  if ($chain =~ /^(PREROUTING|OUTPUT|POSTROUTING)$/)
  {
    $baseChain = $chain;
  }
  else
  {
    # loop through the user defined chains and work our way back till we get to a Builtin chain or -1.
    my $done = 0;
    my $tmpChain = $chain;
    while (!$done)
    {
      if (!exists $self->{userChains}->{$tmpChain})
      {
        die "(NAT->$ruleName) - Error: chain = '$tmpChain' does not exist!\n";
      }
      my $parentChain = $self->{userChains}->{$tmpChain};
      if ($parentChain !~ /^(-1|PREROUTING|OUTPUT|POSTROUTING)$/)
      {
        $tmpChain = $parentChain;
      }
      else
      {
        $baseChain = $parentChain;
        $done = 1;
      }
    }
  }
  return $baseChain;
}

# scalar lookupAlias(int)
# requires: int - interface (i or o value)
# returns: alias associated with that interface uppercased.
sub lookupAlias
{
  my $self = shift;
  my %args = ( int => "", @_ );
  my $int = $args{int};
  my $alias = "";

  (my $int2 = $int) =~ s/^(!?)(.+)$/$2/;  # make sure we don't have a negation taking place as we would never find that interface :)
  my @interfaceInfo = $self->{interfaceObj}->getInfo(int => $int2);
  if (scalar @interfaceInfo == 0)
  {
    die "(lookupAlias) - Error: You have not specified interface '$int2'!  I can not lookup the alias!\nParameters: " . $self->printParameters(%args) . "\n";
  }
  $alias = uc($interfaceInfo[4]);

  return $alias;
}

# scalar lookupDynamicInfo(int)
# requires: int - interface (i or o value)
# returns: @(type, dynamicIPMode) associated with that interface.
sub lookupDynamicInfo
{
  my $self = shift;
  my %args = ( int => "", @_ );
  my $int = $args{int};

  (my $int2 = $int) =~ s/^(!?)(.+)$/$2/;  # make sure we don't have a negation taking place as we would never find that interface :)
  my @interfaceInfo = $self->{interfaceObj}->getInfo(int => $int2);
  if (scalar @interfaceInfo == 0)
  {
    die "(lookupDynamicInfo) - Error: You have not specified interface '$int2'!  I can not lookup the dynamicInfo!\nParameters: " . $self->printParameters(%args) . "\n";
  }

  return ($interfaceInfo[1], $interfaceInfo[5]);
}

# scalar generateRule(%args)
#    This routine will be called by the action methods to do generic validation
#    and to setup and return the first part of the rule upto but not including the -j.
# parameters:
#  i = input interface (only valid in PREROUTING, POSTROUTING)
#  o = output interface (only valid in OUTPUT, POSTROUTING)
#  d = destination address
#  s = source address
#  f = fragment *
#  nf = ! fragment (nf and f can not both be 1) *
#  tcp-flags format is "!? flag,flag,flag flag,flag" (? = optional)
#  syn = --syn *
#  nsyn = ! --syn (syn and nsyn can not both be 1) *
#  mac = -m mac (only valid with chain = PREROUTING or POSTROUTING) *
#    mac-source = !? MAC address to match (mac must = 1)
#  limit = -m limit *
#    limit-rate = --limit rate (limit must be 1)
#    limit-burst (limit must be 1)
#  multiport = -m multiport *
#    source-port (multiport must be 1)
#    destination-port (multiport must be 1)
#    port (multiport must be 1)
#  mark = -m mark *
#    mark-value = --mark value
#  state = -m state *
#    states = --state
#  unclean = -m unclean *
#  tos = -m tos *
#    tos-value = --tos tos
#  chain = name of chain to use (PREROUTING, OUTPUT, POSTROUTING or user defined)**
#  p = protocol (tcp, udp, icmp or "" = all)
#  dport = destination port
#  sport = source port
#  (dport and sport are only valid when p = tcp or udp)
#  int = interface ("" = all interfaces)
#  insert = insert the rule at the specified offset. *
#  append = append the rule to the chain. *  (This is the default action)
#  rule-offset = Offset to insert the rule at.  Defaults to 1.
#  (If append = 1 (by default) and insert = 1 then the insert and
#  rule-offset values will be used, else we append the rule.)
#  log (in all rules except log, generate a log rule before the action) *
#    log-level = syslog level (defaults to notice)
#    log-prefix = string to identify the log entry by.
#    log-tcp-sequence *  (This is a security risk)
#    log-tcp-options *
#    log-ip-options *
#    log-limit = -m limit *
#    log-limit-rate = --limit rate (log-limit must be 1)
#    log-limit-burst (log-limit must be 1)
#  tcp-option (only valid if p = tcp)
#  icmp-type (only valid if p = icmp)
#  ttl = --ttl ttl
sub generateRule
{
  my $self = shift;
  my %args = ( ruleName => "generateRule", @_ );  # no overrides here.
  my $rule = "";

  my $p               = $args{p};
  my $s               = $args{s};
  my $d               = $args{d};
  my $i               = $args{i};
  my $o               = $args{o};
  my $chain           = $args{chain};
  my $f               = $args{f};
  my $nf              = $args{nf};
  my $tcpFlags        = $args{"tcp-flags"};
  my $tcpOption       = $args{"tcp-option"};
  my $icmpType        = $args{"icmp-type"};
  my $syn             = $args{syn};
  my $nsyn            = $args{nsyn};
  my $mac             = $args{mac};
  my $macSource       = $args{"mac-source"};
  my $limit           = $args{limit};
  my $limitRate       = $args{"limit-rate"};
  my $limitBurst      = $args{"limit-burst"};
  my $logLimit        = $args{"log-limit"};
  my $logLimitRate    = $args{"log-limit-rate"};
  my $logLimitBurst   = $args{"log-limit-burst"};
  my $multiport       = $args{multiport};
  my $sourcePort      = $args{"source-port"};
  my $destinationPort = $args{"destination-port"};
  my $port            = $args{port};
  my $mark            = $args{mark};
  my $markValue       = $args{"mark-value"};
  my $state           = $args{state};
  my $states          = $args{states};
  my $unclean         = $args{unclean};
  my $tos             = $args{tos};
  my $tosValue        = $args{"tos-value"};
  my $dport           = $args{dport};
  my $sport           = $args{sport};
  my $insert          = $args{insert};
  my $append          = $args{append};
  my $ruleOffset      = $args{"rule-offset"};
  my $ruleName        = "NAT->" . $args{ruleName};
  my $preComment      = $args{preComment};
  my $log             = $args{log};
  my $ttl             = $args{ttl};

  my $baseChain = "";

  if (length $chain == 0)
  {
    die "($ruleName) - Error: chain must be specified!\nParameters: " . $self->printParameters(%args) . "\n";
  }
  if (length $chain > 30)
  {
    die "($ruleName) - Error: chain = '$chain' is too long!  It can only be 30 characters long.\n";
  }

  $baseChain = $self->getBaseChain(chain => $chain);
  if ($baseChain eq -1)
  {
    #print "($ruleName) - Warn: chain = '$chain' is not referenced!  This rule will be ignored by the kernel!\nParameters: " . $self->printParameters(%args) . "\n";
    $self->{orphanedChains}->{$chain} = (exists $self->{orphanedChains}->{$chain} ? $self->{orphanedChains}->{$chain} + 1 : 1);
  }

  # do quick validation.
  if ($baseChain =~ /^(PREROUTING)$/ && length $o > 0)
  {
    die "($ruleName) - Error: o = '$o' is invalid in baseChain = '$baseChain'!\nParameters: " . $self->printParameters(%args) . "\n";
  }
  if ($baseChain =~ /^(OUTPUT|POSTROUTING)$/ && length $i > 0)
  {
    die "($ruleName) - Error: i = '$i' is invalid in baseChain = '$baseChain'!\nParameters: " . $self->printParameters(%args) . "\n";
  }
  if ($baseChain eq -1 && length $o > 0)
  {
    print "($ruleName) - Warn: chain = '$chain' is not referenced and o = '$o'!  This interface chain pair may not be valid!\nParameters: " . $self->printParameters(%args) . "\n";
  }
  if ($baseChain eq -1 && length $i > 0)
  {
    print "($ruleName) - Warn: chain = '$chain' is not referenced and i = '$i'!  This interface chain pair may not be valid!\nParameters: " . $self->printParameters(%args) . "\n";
  }
  if ($f && (length $sport > 0 || length $dport > 0))
  {
    die "($ruleName) - Error: You can not specify f (fragment) and also specify sport or dport!\nParameters: " . $self->printParameters(%args) . "\n";
  }
  if ($f && $nf)
  {
    die "($ruleName) - Error: You can not specify f (fragment) and nf (! fragment) at the same time!\nParameters: " . $self->printParameters(%args) . "\n";
  }
  if ($syn && $nsyn)
  {
    die "($ruleName) - Error: You can not specify syn and nsyn (! syn) at the same time!\nParameters: " . $self->printParameters(%args) . "\n";
  }
  if ($s =~ /^(!?)(inInterfaceIP|inInterfaceBroadcast)$/ || $d =~ /^(!?)(inInterfaceIP|inInterfaceBroadcast)$/)
  {
    if (length $i == 0)
    {
      die "($ruleName) - Error:  You must specify the input interface (i) if requesting inInterfaceIP/inInterfaceBroadcast for source (s) or destination (d)!\nParameters: " . $self->printParameters(%args) . "\n";
    }
    my $int = $i;
    $int =~ s/^(!?)(.+)$/$2/;  # make sure we don't have a negation taking place as we would never find that interface :)
    my @interfaceInfo = $self->{interfaceObj}->getInfo(int => $int);
    if (scalar @interfaceInfo == 0)
    {
      die "($ruleName) - Error: You have not specified interface '$int'!  I can not do an inInterfaceIP/inInterfaceBroadcast substitution!\nParameters: " . $self->printParameters(%args) . "\n";
    }
  }
  if ($s =~ /^(!?)(outInterfaceIP|outInterfaceBroadcast)$/ || $d =~ /^(!?)(outInterfaceIP|outInterfaceBroadcast)$/)
  {
    if (length $o == 0)
    {
      die "($ruleName) - Error:  You must specify the output interface (o) if requesting outInterfaceIP/outInterfaceBroadcast for source (s) or destination (d)!\nParameters: " . $self->printParameters(%args) . "\n";
    }
    my $int = $o;
    $int =~ s/^(!?)(.+)$/$2/;  # make sure we don't have a negation taking place as we would never find that interface :)
    my @interfaceInfo = $self->{interfaceObj}->getInfo(int => $int);
    if (scalar @interfaceInfo == 0)
    {
      die "($ruleName) - Error: You have not specified interface '$int'!  I can not do an outInterfaceIP/outInterfaceBroadcast substitution!\nParameters: " . $self->printParameters(%args) . "\n";
    }
  }
  if ($multiport && $p !~ /^(tcp|udp|! icmp)$/)
  {
    die "($ruleName) - Error: You can not specify multiport with p = '$p'!\nParameters: " . $self->printParameters(%args) . "\n";
  }
  if (length $icmpType > 0 && $p ne "icmp")
  {
    die "($ruleName) - Error: You can not specify icmp-type with p = '$p'!\nParameters: " . $self->printParameters(%args) . "\n";
  }
  if (length $tcpOption > 0 && $p ne "tcp")
  {
    die "($ruleName) - Error: You can not specify tcp-option with p = '$p'!\nParameters: " . $self->printParameters(%args) . "\n";
  }
  if (length $tcpFlags > 0 && $p ne "tcp")
  {
    die "($ruleName) - Error: You can not specify tcp-flags with p = '$p'!\nParameters: " . $self->printParameters(%args) . "\n";
  }
  if ((length $sport > 0 || length $dport > 0) && $p eq "icmp")
  {
    die "($ruleName) - Error: You can not specify sport or dport with p = '$p'!\nParameters: " . $self->printParameters(%args) . "\n";
  }
  if (length $ttl > 0 && $ttl !~ /^(\d+)$/)
  {
    die "($ruleName) - Error: ttl = '$ttl' is invalid!\nParameters: " . $self->printParameters(%args) . "\n";
  }

  # generate the first part of the rule now.
  $rule = "  ";
  $rule .= $preComment . "  " if (length $preComment > 0);
  $rule .= "\$IPTABLES$TABLE ";
  $rule .= "-A $chain " if (!$insert);
  $rule .= "-I $chain " . ($ruleOffset > 1 ? "$ruleOffset " : "") if ($insert);
  $rule .= "-p $p " if (length $p > 0);

  # get the alias for the specified interface
  my $ialias;
  my $oalias;
  my $itype;
  my $otype;
  my $iDynamicIPMode;
  my $oDynamicIPMode;
  if ($i)
  {
    $ialias = $self->lookupAlias(int => $i);
    ($itype, $iDynamicIPMode) = $self->lookupDynamicInfo(int => $i);
  }
  if ($o)
  {
    $oalias = $self->lookupAlias(int => $o);
    ($otype, $oDynamicIPMode) = $self->lookupDynamicInfo(int => $o);
  }

  $rule .= "-i \${INT_$ialias} " if (length $i > 0);
  $rule .= "-o \${INT_$oalias} " if (length $o > 0);
  my $source = $s;
  if ($source =~ /^(!?inInterfaceIP)$/)
  {
    if ($itype eq "static" || ($itype eq "dynamic" && $iDynamicIPMode eq "lookupIP"))
    {
      $source = "\${$ialias" . "_IP}";
      $source = "! $source" if ($s =~ /^(!)/);
    }
    else
    {
      $source = "";
      $s = "";
    }
  }
  elsif ($source =~ /^(!?outInterfaceIP)$/)
  {
    if ($otype eq "static" || ($otype eq "dynamic" && $oDynamicIPMode eq "lookupIP"))
    {
      $source = "\${$oalias" . "_IP}";
      $source = "! $source" if ($s =~ /^(!)/);
    }
    else
    {
      $source = "";
      $s = "";
    }
  }
  elsif ($source =~ /^(!?inInterfaceBroadcast)$/)
  {
    if ($itype eq "static" || ($itype eq "dynamic" && $iDynamicIPMode eq "lookupIP"))
    {
      (my $int = $i) =~ s/^(!?)(.+)$/$2/;
      $source = "\${$ialias" . "_BROADCAST}";
      $source = "! $source" if ($s =~ /^(!)/);
      $self->{interfaceObj}->usingBroadcast(int => $int);  # set the flag
    }
    else
    {
      $source = "";
      $s = "";
    }
  }
  elsif ($source =~ /^(!?outInterfaceBroadcast)$/)
  {
    if ($otype eq "static" || ($otype eq "dynamic" && $oDynamicIPMode eq "lookupIP"))
    {
      (my $int = $o) =~ s/^(!?)(.+)$/$2/;
      $source = "\${$oalias" . "_BROADCAST}";
      $source = "! $source" if ($s =~ /^(!)/);
      $self->{interfaceObj}->usingBroadcast(int => $int);  # set the flag
    }
    else
    {
      $source = "";
      $s = "";
    }
  }
  $rule .= "-s $source " if (length $s > 0);
  my $destination = $d;
  if ($destination =~ /^(!?inInterfaceIP)$/)
  {
    if ($itype eq "static" || ($itype eq "dynamic" && $iDynamicIPMode eq "lookupIP"))
    {
      $destination = "\${$ialias" . "_IP}";
      $destination = "! $destination" if ($d =~ /^(!)/);
    }
    else
    {
      $destination = "";
      $d = "";
    }
  }
  elsif ($destination =~ /^(!?outInterfaceIP)$/)
  {
    if ($otype eq "static" || ($otype eq "dynamic" && $oDynamicIPMode eq "lookupIP"))
    {
      $destination = "\${$oalias" . "_IP}";
      $destination = "! $destination" if ($d =~ /^(!)/);
    }
    else
    {
      $destination = "";
      $d = "";
    }
  }
  elsif ($destination =~ /^(!?inInterfaceBroadcast)$/)
  {
    if ($itype eq "static" || ($itype eq "dynamic" && $iDynamicIPMode eq "lookupIP"))
    {
      (my $int = $i) =~ s/^(!?)(.+)$/$2/;
      $destination = "\${$ialias" . "_BROADCAST}";
      $destination = "! $destination" if ($d =~ /^(!)/);
      $self->{interfaceObj}->usingBroadcast(int => $int);  # set the flag
    }
    else
    {
      $destination = "";
      $d = "";
    }
  }
  elsif ($destination =~ /^(!?outInterfaceBroadcast)$/)
  {
    if ($otype eq "static" || ($otype eq "dynamic" && $oDynamicIPMode eq "lookupIP"))
    {
      (my $int = $o) =~ s/^(!?)(.+)$/$2/;
      $destination = "\${$oalias" . "_BROADCAST}";
      $destination = "! $destination" if ($d =~ /^(!)/);
      $self->{interfaceObj}->usingBroadcast(int => $int);  # set the flag
    }
    else
    {
      $destination = "";
      $d = "";
    }
  }
  $rule .= "-d $destination " if (length $d > 0);
  $rule .= "-f " if ($f);
  $rule .= "! -f " if ($nf);
  $rule .= "--sport $sport " if (length $sport > 0);
  $rule .= "--dport $dport " if (length $dport > 0);
  $rule .= "--tcp-flags $tcpFlags " if (length $tcpFlags > 0);
  $rule .= "--syn " if ($syn);
  $rule .= "! --syn " if ($nsyn);
  $rule .= "--tcp-option $tcpOption " if (length $tcpOption > 0);
  $rule .= "--icmp-type $icmpType " if (length $icmpType > 0);
  $rule .= "-m mac " if ($mac);
  $rule .= "--mac-source $macSource " if (length $macSource > 0 && $mac);
  $rule .= "-m limit " if (($limit && !$log) || ($logLimit && $log));
  $rule .= "--limit $limitRate " if (length $limitRate > 0 && $limit && !$log);
  $rule .= "--limit $logLimitRate " if (length $logLimitRate > 0 && $logLimit && $log);
  $rule .= "--limit-burst $limitBurst " if (length $limitBurst > 0 && $limit && !$log);
  $rule .= "--limit-burst $logLimitBurst " if (length $logLimitBurst > 0 && $logLimit && $log);
  $rule .= "-m multiport " if ($multiport);
  $rule .= "--source-port $sourcePort " if (length $sourcePort > 0 && $multiport);
  $rule .= "--destination-port $destinationPort " if (length $destinationPort > 0 && $multiport);
  $rule .= "--port $port " if (length $port > 0 && $multiport);
  $rule .= "-m mark " if ($mark);
  $rule .= "--mark $markValue " if (length $markValue > 0 && $mark);
  $rule .= "-m state " if ($state);
  $rule .= "--state $states " if (length $states > 0 && $state);
  $rule .= "-m unclean " if ($unclean);
  $rule .= "-m tos " if ($tos);
  $rule .= "--tos $tosValue " if (length $tosValue > 0 && $tos);
  $rule .= "--ttl $ttl " if (length $ttl > 0);

  return $rule;
}

# void log()
#    A rule will be created to log the specified packets based upon the
#    given parameters.  If nothing is specified, then the rule will be
#    generated for all chains, all protocols, all interfaces.
sub log
{
  my $self = shift;
  my %args = ( p => "", i => "", o => "", chain => "", f => 0, nf => 0, "tcp-flags" => "", syn => 0, nsyn => 0,
               mac => 0, "mac-source" => "", limit => 0, "limit-rate" => "", "limit-burst" => "",
               multiport => 0, "source-port" => "", "destination-port" => "", port => "", mark => 0,
               "mark-value" => "", state => 0, states => "", unclean => 0, tos => 0, "tos-value" => "",
               dport => "", sport => "", insert => 0, append => 1, "rule-offset" => 1, "tcp-option" => "",
               log => 1, "log-level" => "notice", "log-prefix" => $self->{configObj}->{logPrefix},
               "log-tcp-sequence" => 0, "log-tcp-options" => 0, "log-ip-options" => 0, s => "", d => "",
               "icmp-type" => "", preComment => $self->{configObj}->{preComment}, postComment => $self->{configObj}->{postComment},
               ruleMode => "normal", ttl => "", @_ );

  my $log             = $args{log};
  my $logLevel        = $args{"log-level"};
  my $logPrefix       = $args{"log-prefix"};
  my $logTcpSequence  = $args{"log-tcp-sequence"};
  my $logTcpOptions   = $args{"log-tcp-options"};
  my $logIpOptions    = $args{"log-ip-options"};
  my $postComment     = $args{postComment};
  my $ruleMode    = $args{ruleMode};

  die "(NAT->log) - Error:  ruleMode = '$ruleMode' is invalid!\n" if ($ruleMode !~ /^(setup|normal)$/);

  if (length $logPrefix > 29)
  {
    die "(NAT->log) - Error: log-prefix = '$logPrefix' is > 29 characters!\nParameters: " . $self->printParameters(%args) . "\n";
  }

  # get the standard part of the rule
  my $rule = $self->generateRule(%args, ruleName => "log");

  # the action
  $rule .= "-j LOG";

  # any extensions to the action
  $rule .= " --log-level $logLevel" if (length $logLevel > 0);
  $rule .= " --log-prefix \"$logPrefix\"" if (length $logPrefix > 0);
  $rule .= " --log-tcp-sequence" if ($logTcpSequence);
  $rule .= " --log-tcp-options" if ($logTcpOptions);
  $rule .= " --log-ip-options" if ($logIpOptions);
  $rule .= "\n";

  $rule .= "  " . $postComment if (length $postComment > 0);

  my $mode = ($args{p} =~ /^(tcp|udp)$/ && $args{dport} == 53 ? "setup" : $ruleMode);
  $self->appendRule(rule => $rule, mode => "$mode");
}

# void jump(dchain)
#    A rule will be created to jump to the specified chain.
sub jump
{
  my $self = shift;
  my %args = ( p => "", i => "", o => "", chain => "", f => 0, nf => 0, "tcp-flags" => "", syn => 0, nsyn => 0,
               mac => 0, "mac-source" => "", limit => 0, "limit-rate" => "", "limit-burst" => "",
               multiport => 0, "source-port" => "", "destination-port" => "", port => "", mark => 0,
               "mark-value" => "", state => 0, states => "", unclean => 0, tos => 0, "tos-value" => "",
               dport => "", sport => "", insert => 0, append => 1, "rule-offset" => 1, "tcp-option" => "",
               log => 0, "log-level" => "notice", "log-prefix" => $self->{configObj}->{logPrefix},
               "log-tcp-sequence" => 0, "log-tcp-options" => 0, "log-ip-options" => 0, s => "", d => "",
               "icmp-type" => "", preComment => $self->{configObj}->{preComment}, postComment => $self->{configObj}->{postComment},
               dchain => "", ttl => "", @_ );

  my $log         = $args{log};
  my $dchain      = $args{dchain};
  my $chain       = $args{chain};
  my $postComment = $args{postComment};

  # make sure the destination chain is specified and is valid.  It must be a user defined chain.
  if (length $dchain == 0)
  {
    die "(NAT->jump) - Error: dchain must be specified!\nParameters: " . $self->printParameters(%args) . "\n";
  }
  if (!exists $self->{userChains}->{$dchain})
  {
    die "(NAT->jump) - Error: dchain = '$dchain' is not a user defined chain!\nParameters: " . $self->printParameters(%args) . "\n";
  }

  # make sure the source chain is valid.  It can be a builtin chain or user defined.
  if ($chain !~ /^(PREROUTING|OUTPUT|POSTROUTING)$/ && !exists $self->{userChains}->{$chain})
  {
    die "(NAT->jump) - Error: chain = '$chain' is not a builtin or user defined chain!\nParameters: " . $self->printParameters(%args) . "\n";
  }

  # now link the dchain back to the chain we are jumping from.
  $self->{userChains}->{$dchain} = $chain;

  delete $self->{orphanedChains}->{$dchain} if (exists $self->{orphanedChains}->{$dchain});

  $self->log(%args) if ($log);

  $args{log} = 0;  # make sure the limiting code will work correctly.

  # get the standard part of the rule
  my $rule = $self->generateRule(%args, ruleName => "jump");

  # the action
  $rule .= "-j $dchain\n";

  # any extensions to the action

  $rule .= "  " . $postComment if (length $postComment > 0);

  $self->appendRule(rule => $rule, mode => "setup");
}

# void accept()
#    A rule will be created to ACCEPT the specified packet.
sub accept
{
  my $self = shift;
  my %args = ( p => "", i => "", o => "", chain => "", f => 0, nf => 0, "tcp-flags" => "", syn => 0, nsyn => 0,
               mac => 0, "mac-source" => "", limit => 0, "limit-rate" => "", "limit-burst" => "",
               multiport => 0, "source-port" => "", "destination-port" => "", port => "", mark => 0,
               "mark-value" => "", state => 0, states => "", unclean => 0, tos => 0, "tos-value" => "",
               dport => "", sport => "", insert => 0, append => 1, "rule-offset" => 1, "tcp-option" => "",
               log => 0, "log-level" => "notice", "log-prefix" => $self->{configObj}->{logPrefix},
               "log-tcp-sequence" => 0, "log-tcp-options" => 0, "log-ip-options" => 0, s => "", d => "",
               "icmp-type" => "", preComment => $self->{configObj}->{preComment}, postComment => $self->{configObj}->{postComment},
               ruleMode => "normal", ttl => "", @_ );

  my $log         = $args{log};
  my $postComment = $args{postComment};
  my $ruleMode    = $args{ruleMode};

  die "(NAT->accept) - Error:  ruleMode = '$ruleMode' is invalid!\n" if ($ruleMode !~ /^(setup|normal)$/);

  $self->log(%args) if ($log);

  $args{log} = 0;  # make sure the limiting code will work correctly.

  # get the standard part of the rule
  my $rule = $self->generateRule(%args, ruleName => "accept");

  # the action
  $rule .= "-j ACCEPT\n";

  # any extensions to the action

  $rule .= "  " . $postComment if (length $postComment > 0);

  my $mode = ($args{p} =~ /^(tcp|udp)$/ && $args{dport} == 53 ? "setup" : $ruleMode);
  $self->appendRule(rule => $rule, mode => "$mode");
}

# void drop()
#    A rule will be created to DROP the specified packet.
sub drop
{
  my $self = shift;
  my %args = ( p => "", i => "", o => "", chain => "", f => 0, nf => 0, "tcp-flags" => "", syn => 0, nsyn => 0,
               mac => 0, "mac-source" => "", limit => 0, "limit-rate" => "", "limit-burst" => "",
               multiport => 0, "source-port" => "", "destination-port" => "", port => "", mark => 0,
               "mark-value" => "", state => 0, states => "", unclean => 0, tos => 0, "tos-value" => "",
               dport => "", sport => "", insert => 0, append => 1, "rule-offset" => 1, "tcp-option" => "",
               log => 0, "log-level" => "notice", "log-prefix" => $self->{configObj}->{logPrefix},
               "log-tcp-sequence" => 0, "log-tcp-options" => 0, "log-ip-options" => 0, s => "", d => "",
               "icmp-type" => "", preComment => $self->{configObj}->{preComment}, postComment => $self->{configObj}->{postComment},
               ruleMode => "normal", ttl => "", @_ );

  my $log         = $args{log};
  my $postComment = $args{postComment};
  my $ruleMode    = $args{ruleMode};

  die "(NAT->drop) - Error:  ruleMode = '$ruleMode' is invalid!\n" if ($ruleMode !~ /^(setup|normal)$/);

  $self->log(%args) if ($log);

  $args{log} = 0;  # make sure the limiting code will work correctly.

  # get the standard part of the rule
  my $rule = $self->generateRule(%args, ruleName => "drop");

  # the action
  $rule .= "-j DROP\n";

  # any extensions to the action

  $rule .= "  " . $postComment if (length $postComment > 0);

  my $mode = ($args{p} =~ /^(tcp|udp)$/ && $args{dport} == 53 ? "setup" : $ruleMode);
  $self->appendRule(rule => $rule, mode => "$mode");
}

# void return()
#    A rule will be created to RETURN to the calling chain.
sub return
{
  my $self = shift;
  my %args = ( p => "", i => "", o => "", chain => "", f => 0, nf => 0, "tcp-flags" => "", syn => 0, nsyn => 0,
               mac => 0, "mac-source" => "", limit => 0, "limit-rate" => "", "limit-burst" => "",
               multiport => 0, "source-port" => "", "destination-port" => "", port => "", mark => 0,
               "mark-value" => "", state => 0, states => "", unclean => 0, tos => 0, "tos-value" => "",
               dport => "", sport => "", insert => 0, append => 1, "rule-offset" => 1, "tcp-option" => "",
               log => 0, "log-level" => "notice", "log-prefix" => $self->{configObj}->{logPrefix},
               "log-tcp-sequence" => 0, "log-tcp-options" => 0, "log-ip-options" => 0, s => "", d => "",
               "icmp-type" => "", preComment => $self->{configObj}->{preComment}, postComment => $self->{configObj}->{postComment},
               ttl => "", @_ );

  my $log         = $args{log};
  my $postComment = $args{postComment};

  $self->log(%args) if ($log);

  $args{log} = 0;  # make sure the limiting code will work correctly.

  # get the standard part of the rule
  my $rule = $self->generateRule(%args, ruleName => "return");

  # the action
  $rule .= "-j RETURN\n";

  # any extensions to the action

  $rule .= "  " . $postComment if (length $postComment > 0);

  $self->appendRule(rule => $rule, mode => "setup");
}

# void queue()
#    A rule will be created to QUEUE the packet to user space.
sub queue
{
  my $self = shift;
  my %args = ( p => "", i => "", o => "", chain => "", f => 0, nf => 0, "tcp-flags" => "", syn => 0, nsyn => 0,
               mac => 0, "mac-source" => "", limit => 0, "limit-rate" => "", "limit-burst" => "",
               multiport => 0, "source-port" => "", "destination-port" => "", port => "", mark => 0,
               "mark-value" => "", state => 0, states => "", unclean => 0, tos => 0, "tos-value" => "",
               dport => "", sport => "", insert => 0, append => 1, "rule-offset" => 1, "tcp-option" => "",
               log => 0, "log-level" => "notice", "log-prefix" => $self->{configObj}->{logPrefix},
               "log-tcp-sequence" => 0, "log-tcp-options" => 0, "log-ip-options" => 0, s => "", d => "",
               "icmp-type" => "", preComment => $self->{configObj}->{preComment}, postComment => $self->{configObj}->{postComment},
               ttl => "", @_ );

  my $log         = $args{log};
  my $postComment = $args{postComment};

  $self->log(%args) if ($log);

  $args{log} = 0;  # make sure the limiting code will work correctly.

  # get the standard part of the rule
  my $rule = $self->generateRule(%args, ruleName => "queue");

  # the action
  $rule .= "-j QUEUE\n";

  # any extensions to the action

  $rule .= "  " . $postComment if (length $postComment > 0);

  $self->appendRule(rule => $rule, mode => "setup");
}

# void snat(to-source)
#    A rule will be created to SNAT the specified packet to appear to come
#    from the specified source.
sub snat
{
  my $self = shift;
  my %args = ( p => "", i => "", o => "", chain => "", f => 0, nf => 0, "tcp-flags" => "", syn => 0, nsyn => 0,
               mac => 0, "mac-source" => "", limit => 0, "limit-rate" => "", "limit-burst" => "",
               multiport => 0, "source-port" => "", "destination-port" => "", port => "", mark => 0,
               "mark-value" => "", state => 0, states => "", unclean => 0, tos => 0, "tos-value" => "",
               dport => "", sport => "", insert => 0, append => 1, "rule-offset" => 1, "tcp-option" => "",
               log => 0, "log-level" => "notice", "log-prefix" => $self->{configObj}->{logPrefix},
               "log-tcp-sequence" => 0, "log-tcp-options" => 0, "log-ip-options" => 0, s => "", d => "",
               "icmp-type" => "", preComment => $self->{configObj}->{preComment}, postComment => $self->{configObj}->{postComment},
               "to-source" => "", ruleMode => "normal", ttl => "", @_ );

  my $log         = $args{log};
  my $postComment = $args{postComment};
  my $toSource    = $args{"to-source"};
  my $chain       = $args{chain};
  my $o           = $args{o};
  my $ruleMode    = $args{ruleMode};

  die "(NAT->snat) - Error:  ruleMode = '$ruleMode' is invalid!\n" if ($ruleMode !~ /^(setup|normal)$/);

  if (length $toSource == 0)
  {
    die "(NAT->snat) - Error:  You must specify to-source!\nParameters: " . $self->printParameters(%args) . "\n";
  }
  my $baseChain = $self->getBaseChain(chain => $chain, ruleName => "snat");
  if ($baseChain !~ /^(POSTROUTING)$/)
  {
    die "(NAT->snat) - Error:  Your chain is not derived from POSTROUTING!\nParameters: " . $self->printParameters(%args) . "\n";
  }
  if ($toSource eq "outInterfaceIP")
  {
    if (length $o == 0)
    {
      die "(NAT->snat) - Error:  You must specify the output interface (o) if requesting outInterfaceIP for to-source!\nParameters: " . $self->printParameters(%args) . "\n";
    }
    # get the alias for the specified interface
    my $oalias = $self->lookupAlias(int => $o);
  }

  $self->log(%args) if ($log);

  $args{log} = 0;  # make sure the limiting code will work correctly.

  # get the standard part of the rule
  my $rule = $self->generateRule(%args, ruleName => "snat");

  # the action
  $rule .= "-j SNAT";

  # any extensions to the action
  my $source = $toSource;
  if ($source =~ /^(!?outInterfaceIP)$/)
  {
    my $oalias = $self->lookupAlias(int => $o);
    $source = "\${$oalias" . "_IP}";
  }
  $rule .= " --to-source $source";

  $rule .= "\n";

  $rule .= "  " . $postComment if (length $postComment > 0);

  my $mode = ($args{p} =~ /^(tcp|udp)$/ && $args{dport} == 53 ? "setup" : $ruleMode);
  $self->appendRule(rule => $rule, mode => "$mode");
}

# void dnat(to-destination)
#    A rule will be created to DNAT the specified packet to the specified
#    destination.
sub dnat
{
  my $self = shift;
  my %args = ( p => "", i => "", o => "", chain => "", f => 0, nf => 0, "tcp-flags" => "", syn => 0, nsyn => 0,
               mac => 0, "mac-source" => "", limit => 0, "limit-rate" => "", "limit-burst" => "",
               multiport => 0, "source-port" => "", "destination-port" => "", port => "", mark => 0,
               "mark-value" => "", state => 0, states => "", unclean => 0, tos => 0, "tos-value" => "",
               dport => "", sport => "", insert => 0, append => 1, "rule-offset" => 1, "tcp-option" => "",
               log => 0, "log-level" => "notice", "log-prefix" => $self->{configObj}->{logPrefix},
               "log-tcp-sequence" => 0, "log-tcp-options" => 0, "log-ip-options" => 0, s => "", d => "",
               "icmp-type" => "", preComment => $self->{configObj}->{preComment}, postComment => $self->{configObj}->{postComment},
               "to-destination" => "", ruleMode => "normal", ttl => "", @_ );

  my $log           = $args{log};
  my $postComment   = $args{postComment};
  my $toDestination = $args{"to-destination"};
  my $chain         = $args{chain};
  my $ruleMode    = $args{ruleMode};

  die "(NAT->dnat) - Error:  ruleMode = '$ruleMode' is invalid!\n" if ($ruleMode !~ /^(setup|normal)$/);

  if (length $toDestination == 0)
  {
    die "(NAT->dnat) - Error:  You must specify to-destination!\nParameters: " . $self->printParameters(%args) . "\n";
  }
  my $baseChain = $self->getBaseChain(chain => $chain, ruleName => "dnat");
  if ($baseChain !~ /^(PREROUTING|OUTPUT)$/)
  {
    die "(NAT->dnat) - Error:  Your chain is not derived from PREROUTING or OUTPUT!\nParameters: " . $self->printParameters(%args) . "\n";
  }

  $self->log(%args) if ($log);

  $args{log} = 0;  # make sure the limiting code will work correctly.

  # get the standard part of the rule
  my $rule = $self->generateRule(%args, ruleName => "dnat");

  # the action
  $rule .= "-j DNAT";

  # any extensions to the action
  $rule .= " --to-destination $toDestination";

  $rule .= "\n";

  $rule .= "  " . $postComment if (length $postComment > 0);

  my $mode = ($args{p} =~ /^(tcp|udp)$/ && $args{dport} == 53 ? "setup" : $ruleMode);
  $self->appendRule(rule => $rule, mode => "$mode");
}

# void masquerade(to-ports)
#    A rule will be created to MASQUERADE the specified packet to appear
#    to come from the machine.  You can specify the to-ports value to
#    override the SNAT source port selection heuristics, but you must
#    specify p = tcp or udp.
sub masquerade
{
  my $self = shift;
  my %args = ( p => "", i => "", o => "", chain => "", f => 0, nf => 0, "tcp-flags" => "", syn => 0, nsyn => 0,
               mac => 0, "mac-source" => "", limit => 0, "limit-rate" => "", "limit-burst" => "",
               multiport => 0, "source-port" => "", "destination-port" => "", port => "", mark => 0,
               "mark-value" => "", state => 0, states => "", unclean => 0, tos => 0, "tos-value" => "",
               dport => "", sport => "", insert => 0, append => 1, "rule-offset" => 1, "tcp-option" => "",
               log => 0, "log-level" => "notice", "log-prefix" => $self->{configObj}->{logPrefix},
               "log-tcp-sequence" => 0, "log-tcp-options" => 0, "log-ip-options" => 0, s => "", d => "",
               "icmp-type" => "", preComment => $self->{configObj}->{preComment}, postComment => $self->{configObj}->{postComment},
               "to-ports" => "", ruleMode => "normal", ttl => "", @_ );

  my $log         = $args{log};
  my $postComment = $args{postComment};
  my $toPorts     = $args{"to-ports"};
  my $chain       = $args{chain};
  my $o           = $args{o};
  my $p           = $args{p};
  my $ruleMode    = $args{ruleMode};

  die "(NAT->masquerade) - Error:  ruleMode = '$ruleMode' is invalid!\n" if ($ruleMode !~ /^(setup|normal)$/);

  my $baseChain = $self->getBaseChain(chain => $chain, ruleName => "masquerade");
  if ($baseChain !~ /^(POSTROUTING)$/)
  {
    die "(NAT->masquerade) - Error:  Your chain is not derived from POSTROUTING!\nParameters: " . $self->printParameters(%args) . "\n";
  }
  if (length $o == 0)
  {
    die "(NAT->masquerade) - Error:  You must specify the output interface (o)!\nParameters: " . $self->printParameters(%args) . "\n";
  }
  if (length $toPorts > 0 && $p !~ /^(tcp|udp)$/)
  {
    die "(NAT->masquerade) - Error:  You can not specify to-ports if p = '$p'!\nParameters: " . $self->printParameters(%args) . "\n";
  }

  $self->log(%args) if ($log);

  $args{log} = 0;  # make sure the limiting code will work correctly.

  # get the standard part of the rule
  my $rule = $self->generateRule(%args, ruleName => "masquerade");

  # the action
  $rule .= "-j MASQUERADE";

  # any extensions to the action
  $rule .= " --to-ports $toPorts" if (length $toPorts > 0);

  $rule .= "\n";

  $rule .= "  " . $postComment if (length $postComment > 0);

  my $mode = ($args{p} =~ /^(tcp|udp)$/ && $args{dport} == 53 ? "setup" : $ruleMode);
  $self->appendRule(rule => $rule, mode => "$mode");
}

# void redirect(to-ports)
#    A rule will be created to REDIRECT the specified packet to the local
#    machine.  If you specify the to-ports value, you can modify where
#    the packet is sent.  You must specify p = tcp or udp to use to-ports.
sub redirect
{
  my $self = shift;
  my %args = ( p => "", i => "", o => "", chain => "", f => 0, nf => 0, "tcp-flags" => "", syn => 0, nsyn => 0,
               mac => 0, "mac-source" => "", limit => 0, "limit-rate" => "", "limit-burst" => "",
               multiport => 0, "source-port" => "", "destination-port" => "", port => "", mark => 0,
               "mark-value" => "", state => 0, states => "", unclean => 0, tos => 0, "tos-value" => "",
               dport => "", sport => "", insert => 0, append => 1, "rule-offset" => 1, "tcp-option" => "",
               log => 0, "log-level" => "notice", "log-prefix" => $self->{configObj}->{logPrefix},
               "log-tcp-sequence" => 0, "log-tcp-options" => 0, "log-ip-options" => 0, s => "", d => "",
               "icmp-type" => "", preComment => $self->{configObj}->{preComment}, postComment => $self->{configObj}->{postComment},
               "to-ports" => "", ruleMode => "normal", ttl => "", @_ );

  my $log         = $args{log};
  my $postComment = $args{postComment};
  my $toPorts     = $args{"to-ports"};
  my $chain       = $args{chain};
  my $p           = $args{p};
  my $ruleMode    = $args{ruleMode};

  die "(NAT->redirect) - Error:  ruleMode = '$ruleMode' is invalid!\n" if ($ruleMode !~ /^(setup|normal)$/);

  my $baseChain = $self->getBaseChain(chain => $chain, ruleName => "redirect");
  if ($baseChain !~ /^(PREROUTING|OUTPUT)$/)
  {
    die "(NAT->redirect) - Error:  Your chain is not derived from PREROUTING or OUTPUT!\nParameters: " . $self->printParameters(%args) . "\n";
  }
  if (length $toPorts > 0 && $p !~ /^(tcp|udp)$/)
  {
    die "(NAT->redirect) - Error:  You can not specify to-ports if p = '$p'!\nParameters: " . $self->printParameters(%args) . "\n";
  }

  $self->log(%args) if ($log);

  $args{log} = 0;  # make sure the limiting code will work correctly.

  # get the standard part of the rule
  my $rule = $self->generateRule(%args, ruleName => "redirect");

  # the action
  $rule .= "-j REDIRECT";

  # any extensions to the action
  $rule .= " --to-ports $toPorts" if (length $toPorts > 0);

  $rule .= "\n";

  $rule .= "  " . $postComment if (length $postComment > 0);

  my $mode = ($args{p} =~ /^(tcp|udp)$/ && $args{dport} == 53 ? "setup" : $ruleMode);
  $self->appendRule(rule => $rule, mode => "$mode");
}

# scalar display(mode)
#    This routine will write the rules to a string.  This is called
#    by the generator script.
#    If mode = "setup" then output the setup rules, else output normal rules
sub display
{
  my $self = shift;
  my %args = ( mode => "normal", @_ );
  my $mode = $args{mode};
  my $result = "";

  if ($mode !~ /^(setup|normal)$/)
  {
    die "(NAT->display) - Error: mode = '$mode' is invalid!\n";
  }

  if ($mode eq "normal")
  {
    $result .= "\n  #\n  # NAT Module Rules.\n  #\n\n";

    # now output all rules/actions we generated.
    foreach my $rule (@{$self->{rules}})
    {
      $result .= $rule;
    }

    if (scalar(keys %{$self->{orphanedChains}}) > 0)
    {
      print "\nThe following userDefined Chains in Table NAT were used but never JUMPed to!\n";
      print "Name\t\tRef Count\n";
      print "----\t\t---------\n";
      foreach my $chain (keys %{$self->{orphanedChains}})
      {
        print "$chain\t\t$self->{orphanedChains}->{$chain}\n";
      }
      print "\n";
    }
  }
  else
  {
    $result .= "\n  #\n  # NAT Module Setup Rules.\n  #\n\n";

    # now output all rules/actions we generated.
    foreach my $rule (@{$self->{setupRules}})
    {
      $result .= $rule;
    }
  }

  return $result;
}

# void appendRule(rule, mode)
#    This routine will append the rule to the end of the rules array.
#    This is called internally by all the rule generating routines in this
#    module.
#    If mode = "setup" then it goes on the setupRules else on the rules array.
sub appendRule
{
  my $self = shift;
  my %args = ( rule => "", mode => "normal", @_ );
  my $rule = $args{rule};
  my $mode = $args{mode};

  if (length $rule == 0)
  {
    die "(NAT->appendRule) - Error:  rule must be defined!\n";
  }
  if ($mode !~ /^(setup|normal)$/)
  {
    die "(NAT->appendRule) - Error:  mode = '$mode' is invalid!\n";
  }

  if ($mode eq "normal")
  {
    push @{$self->{rules}}, $rule;  # append to end of rules.
  }
  else
  {
    push @{$self->{setupRules}}, $rule;  # append to end of setupRules.
  }
}

# scalar printParameters(%args)
#    This routine takes the hash of arguments that were passed into the calling method
#    and returns them in a string, comma seperated.  Only those parameters that are
#    actually specified (not 0 or non-empty strings) will be pulled out.
sub printParameters
{
  my $self = shift;
  my %args = ( @_ );  # no overloading here.
  my $result = "";

  foreach my $arg (keys %args)
  {
    if ($arg =~ /^(nf|f|syn|nsyn|mac|limit|multiport|mark|state|unclean|tos|insert|log-tcp-sequence|log-tcp-options|log-ip-options|log|log-limit)$/)
    {
      $result .= (length $result > 0 ? ", " : "") . $arg . " => " . $args{$arg} if ($args{$arg} == 1);
    }
    elsif ($arg eq "append" && $args{$arg} == 0)
    {
      $result .= (length $result > 0 ? ", " : "") . $arg . " => " . $args{$arg};
    }
    elsif ($arg eq "rule-offset" && $args{$arg} != 1)
    {
      $result .= (length $result > 0 ? ", " : "") . $arg . " => " . $args{$arg};
    }
    elsif ($arg eq "chain")
    {
      $result .= (length $result > 0 ? ", " : "") . $arg . " => " . $args{$arg};
    }
    elsif ($arg eq "log-level" && $args{$arg} ne "notice")
    {
      $result .= (length $result > 0 ? ", " : "") . $arg . " => " . $args{$arg};
    }
    elsif ($arg eq "log-prefix" && $args{$arg} ne $self->{configObj}->{logPrefix})
    {
      $result .= (length $result > 0 ? ", " : "") . $arg . " => " . $args{$arg};
    }
    elsif ($arg eq "preComment" && $args{$arg} ne $self->{configObj}->{preComment})
    {
      $result .= (length $result > 0 ? ", " : "") . $arg . " => " . $args{$arg};
    }
    elsif ($arg eq "postComment" && $args{$arg} ne $self->{configObj}->{postComment})
    {
      $result .= (length $result > 0 ? ", " : "") . $arg . " => " . $args{$arg};
    }
    elsif (length $args{$arg} > 0)
    {
      next if ($arg =~ /^(append|rule-offset|chain|log-level|log-prefix|preComment|postComment)$/);
      $result .= (length $result > 0 ? ", " : "") . $arg . " => " . $args{$arg};
    }
  }

  return $result;
}

1;
__END__

=head1 NAME

NAT - The PCXFireWall NAT Table Module.

=head1 SYNOPSIS

  use PCXFireWall::NAT;
  my $obj = PCXFireWall::NAT->new(configObj => $configObj,
            interfaceObj => $interfaceObj);


=head1 DESCRIPTION

NAT is the Module for the PCXFireWall representing the NAT Table.

=head1 FUNCTIONS

  scalar new(configObj, interfaceObj)
    Creates a new instance of the PCXFireWall::NAT
    object.  This is called by the generator script.  configObj
    allows us to be aware of the config changes the user makes in
    regards to the installPath, comments, etc.  interfaceObj allows
    us to be aware of the network interfaces the user defines.

  void policy(chain, policy, preComment, postComment)
    Sets the default policy of chain to policy.  Valid values are
    PREROUTING, OUTPUT and POSTROUTING for chain and ACCEPT or DENY for policy.

  void createChain(name, preComment, postComment)
    Makes an entry in userChains for the specified chain and then makes
    an entry in rules for the command to create the chain.

  void flushChain(name, preComment, postComment)
    name can be 'All' which will empty all user defined chains.
    A rule will be added which would flush the specified chain.

  void deleteChain(name, preComment, postComment)
    name can be 'All' which will delete all user defined chains.
    The chain(s) will be removed from userChains (if present) and then
    a rule(s) will be added which would delete the chain.  The chain(s)
    will have rules created to flush them before we delete it so that
    it will always work.  Thus you don't need to call flushChain unless
    you are not planning on deleting the chain.



  Valid parameters for the "rules" methods.  If string is empty, then
  that parameter is left out of the generated rule and iptables will
  use whatever default it deems fit.  Items marked with a * are boolean
  values and take 1 or 0.  If 0, then they are ignored (they default to
  being 0).  Any parameters marked with a ** are required.

  i = input interface (only valid in PREROUTING)
  o = output interface (only valid in OUTPUT or POSTROUTING)
  s = source address (if value is inInterfaceIP then I will substitue
                      the IP address of i, else if value is
                      outInterfaceIP then I will substitue the IP
                      address of o, else we use the specified value.
                      If value is !inInterfaceIP or !outInterfaceIP,
                      then the resulting IP Address is prefixed with
                      a ! space.
                      inInterfaceBroadcast and outInterfaceBroadcast
                      will use the value of the shell variable
                      <interface>BROADCAST for the source address.)
  d = destination address (same as s for inInterfaceIP/outInterfaceIP
                         and inInterfaceBroadcast/outInterfaceBroadcast)
  f = fragment *
  nf = ! fragment (nf and f can not both be 1) *
  tcp-flags format is "!? flag,flag,flag flag,flag" (? = optional)
  syn = --syn *
  nsyn = ! --syn (syn and nsyn can not both be 1) *
  mac = -m mac (only valid with chain = PREROUTING or POSTROUTING) *
    mac-source = !? MAC address to match (mac must = 1)
  limit = -m limit *
    limit-rate = --limit rate (limit must be 1)
    limit-burst (limit must be 1)
  multiport = -m multiport *
    source-port (multiport must be 1)
    destination-port (multiport must be 1)
    port (multiport must be 1)
  mark = -m mark *
    mark-value = --mark value
  state = -m state *
    states = --state
  unclean = -m unclean *
  tos = -m tos *
    tos-value = --tos tos
  chain = name of chain to use (PREROUTING, OUTPUT, POSTROUTING or user defined)**
  p = protocol (tcp, udp, icmp or "" = all)
  dport = destination port
  sport = source port
  (dport and sport are only valid when p = tcp or udp)
  int = interface ("" = all interfaces)
  insert = insert the rule at the specified offset. *
  append = append the rule to the chain. *  (This is the default action)
  rule-offset = Offset to insert the rule at.  Defaults to 1.
  (If append = 1 (by default) and insert = 1 then the insert and
  rule-offset values will be used, else we append the rule.)
  log (in all rules except log, generate a log rule before the action) *
    log-level = syslog level (defaults to notice)
    log-prefix = string to identify the log entry by.
    log-tcp-sequence *  (This is a security risk)
    log-tcp-options *
    log-ip-options *
    (to use any of these options, log must be = 1 or in rule log)
    log-limit = -m limit *
    log-limit-rate = --limit rate (log-limit must be 1)
    log-limit-burst (log-limit must be 1)
  tcp-option (only valid with p = tcp)
  icmp-type (only valid with p = icmp)
  ttl = --ttl ttl

  preComment - text before the rule.
  postComment - text after the rule.

  ruleMode - (setup or normal) Allows you to specify which array the rule
              should be placed in so that it gets output correctly.

  See the man page for iptables for valid combinations.


  NOTE: When using any of the following methods and you specify log => 1,
    you must also specify log-limit, log-limit-rate, log-limit-burst
    as seperate from limit, limit-rate and limit-burst if you want to
    limit the log rule and the parent rule at different rates.  If you
    only specify limit, limit-rate and limit-burst then only the parent
    rule will be limited, the log rule will not be limited.  Thus if you
    want to limit both the log and the parent rule you must specify both
    limit, log-limit, limit-rate, and log-limit-rate, etc.  This allows
    you to limit the logging but not the rule itself, and vice a versa.

    The parent rule will set log = 0 to make sure that any limiting you
    wanted to be done will take place.


  void log()
    A rule will be created to log the specified packets based upon the
    given parameters.  If nothing is specified, then the rule will be
    generated for all chains, all protocols, all interfaces.

  void accept()
    A rule will be created to ACCEPT the specified packets based upon
    the given parameters.  If nothing is specified, then the rule will be
    generated for all chains, all protocols, all interfaces.

  void drop()
    A rule will be created to DROP the specified packets based upon
    the given parameters.  If nothing is specified, then the rule will be
    generated for all chains, all protocols, all interfaces.

  void return()
     A rule will be created to RETURN to the calling chain.

  void queue()
     A rule will be created to QUEUE the packet to user space.

  void jump(dchain)
    A rule will be created to JUMP to the specified chain based upon
    the given parameters.  chain must be specified.  dchain is the
    destination chain to jump to and must also be specified.  dchain can
    only be a user defined chain.

  void snat(to-source)
    A rule will be created to SNAT the specified packet to appear to come
    from the specified source.

  void dnat(to-destination)
    A rule will be created to DNAT the specified packet to the specified
    destination.

  void masquerade(to-ports)
    A rule will be created to MASQUERADE the specified packet to appear
    to come from the machine.  You can specify the to-ports value to
    override the SNAT source port selection heuristics, but you must
    specify p = tcp or udp.

  void redirect(to-ports)
    A rule will be created to REDIRECT the specified packet to the local
    machine.  If you specify the to-ports value, you can modify where
    the packet is sent.  You must specify p = tcp or udp to use to-ports.


  Helper methods for /proc entries.

    You must use the Proc modules methods.

  Internally used methods.

  scalar display(mode)
    This routine will write the rules to a string.  This is called
    by the generator script.  If mode = 'normal' then we display all rules
    from the rules array, else if mode = 'setup' then we display all
    rules from the setupRules array which contains chain creation,
    flushing, etc. and any rules that are detected to relate to DNS.

  void appendRule(rule, mode)
    This routine will append the rule to the end of the rules array.
    This is called internally by all the rule generating routines in this
    module.  If mode = 'normal' then the rule is appended to the rules
    array else if mode = 'setup' then the rule is appened to the
    setupRules array.

  scalar lookupAlias(int)
    requires: int - interface (i or o value)
    returns: alias associated with that interface uppercased.

  scalar lookupDynamicInfo(int)
    requires: int - interface (i or o value)
    returns: @(type, dynamicIPMode) associated with that interface.

=head1 VARIABLES

  userChains - The chains that the user creates.  Type is hash.

  rules - All rules that will be generated are stored here.
          Type is array.

  setupRules - All setup and dns related rules will be stored here.
               Type is array.

  NOTE:  All data fields are accessible by specifying the object
         and pointing to the data member to be modified on the
         left-hand side of the assignment.
         Ex.  $obj->variable($newValue); or $value = $obj->variable;

=head1 AUTHOR

Xperience, Inc. (mailto:admin at pcxperience.com)

=head1 SEE ALSO

perl(1)

=cut
