# Network.pm - Interface module for defining network host entries for generator (IPTables Firewall Configuration).
# Created by James A. Pattie.  Copyright (c) 2001-2004, Xperience, Inc.

package PCXFireWall::Network;

use strict;
use vars qw ($AUTOLOAD @ISA @EXPORT $VERSION);

require Exporter;
@ISA = qw(Exporter AutoLoader);
@EXPORT = qw();

$VERSION = "2.23";

sub new
{
  my $that = shift;
  my $class = ref($that) || $that;
  my $self = bless {}, $class;
  $self->{entries} = {};
  $self->{entries}->{'_order'} = [];  # the order in which to output.

  return $self;
}

sub AUTOLOAD
{
  my $self = shift;
  my $type = ref($self) || die "$self is not an object";
  my $name = $AUTOLOAD;
  $name =~ s/.*://;	# strip fully-qualified portion
  unless (exists $self->{$name})
  {
    die "Can't access `$name' field in object of class $type";
  }
  if (@_)
  {
    return $self->{$name} = shift;
  }
  else
  {
    return $self->{$name};
  }
}

# register(index, host, comment)
#  required: host - network, MAC address, ip address
#            index - the index of the network entry to create.
#            comment - user defined comment for this entry
#  optional:
#  summary:  This makes an entry in the entries hash of the form
#            NETWORK_{index} = host
#            with the comment displayed above it.
sub register
{
  my $self = shift;
  my %args = ( index => "", host => "", comment => "", @_ );
  my $index = $args{index};
  my $host = $args{host};
  my $comment = $args{comment};
  my $type = "network";
  my $errStr = "PCXFireWall::Network->register()  - Error!\n";

  if ($index !~ /^(\d+)$/)
  {
    die $errStr . "index = '$index' is invalid!\n";
  }
  if (length $host == 0)
  {
    die $errStr . "host must be defined!\n";
  }

  my $entry = uc($type) . "_$index";

  if (exists $self->{entries}->{$entry})
  {
    die $errStr . "An entry with index = '$index' already defined for type = '$type'!\n";
  }

  my $code = "";

  if ($comment)
  {
    $code .= <<"END_OF_CODE";
  # $comment
END_OF_CODE
  }

  $code .= <<"END_OF_CODE";
  $entry="$host"
END_OF_CODE

  push @{$self->{entries}->{'_order'}}, $entry;
  $self->{entries}->{$entry} = $code;
}

# scalar display()
#   This routine will write the rules to a string.  This is
#   called by the generator script.
sub display
{
  my $self = shift;
  my %args = ( @_ );
  my $result = "";

  $result .= "\n  #\n  # User defined Networks and Hosts.\n  #\n\n";

  # now output all entries we generated.
  foreach my $entry (@{$self->{entries}->{'_order'}})
  {
    $result .= $self->{entries}->{$entry};
  }
  
  return $result;
}

1;
__END__

=head1 NAME

Network - The PCXFireWall Network Host definition interface.

=head1 SYNOPSIS

  use PCXFireWall::Network;
  my $networks = PCXFireWall::Network->new;

  # specify the Internal network
  $networks->register(index => "0", host => "192.168.0.0/24",
                     comment => "Internal Network");

=head1 DESCRIPTION

Network is the netfilter Network Definition Object for the PCXFireWall.

=head1 FUNCTIONS

  scalar new(void)
    Creates a new instance of the PCXFireWall::Network object.
    Initializes the entries hash.

  scalar display()
    This routine will write the rules to a string.  This is
    called by the generator script.

  void register(index, host, comment)
   required: host - network, MAC address, ip address
             index - the index of the network entry to create.
             comment - user defined comment for this entry
    optional:
    summary:  This makes an entry in the entries hash  of the form
              NETWORK_{index} = {host}

=head1 VARIABLES

  entries - hash of network host entries that need to be available to the
            shell script as NETWORK_# (where # is the index
            number 0, 1, 2, etc.)

  NOTE:  All data fields are accessible by specifying the object
         and pointing to the data member to be modified on the
         left-hand side of the assignment.
         Ex.  $obj->variable($newValue); or $value = $obj->variable;

=head1 AUTHOR

Xperience, Inc. (mailto:admin at pcxperience.com)

=head1 SEE ALSO

perl(1)

=cut
